__all__ = (
    'create_attachment_input', 'create_attachment_media', 'create_button', 'create_channel_select', 'create_container',
    'create_label', 'create_media_gallery', 'create_mentionable_select', 'create_role_select', 'create_row',
    'create_section', 'create_separator', 'create_string_select', 'create_text_display', 'create_text_input',
    'create_thumbnail_media', 'create_user_select'
)

from .component import Component, ComponentType


def create_button(
    label = ...,
    emoji = ...,
    *,
    custom_id = ...,
    enabled = ...,
    sku_id = ...,
    style = ...,
    url = ...,
):
    """
    Creates a new button component.
    
    Parameters
    -----------
    label : `None | str`, Optional
        Label of the component.
    
    emoji : ``None | Emoji``, Optional
        Emoji of the button if applicable.
    
    custom_id : `None | str`, Optional (Keyword only)
        Custom identifier to detect which button was clicked by the user.
        
        > Mutually exclusive with the `url` and `sku_id` fields.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    sku_id : ``int | SKU``, Optional (Keyword only)
        Purchasable stock keeping unit identifier.
        
        > Mutually exclusive with the `custom_id` and `url` fields.
    
    style : `None`, ``int | ButtonStyle``, Optional (Keyword only)
        The button's style.
    
    url : `None | str`, Optional (Keyword only)
        Url to redirect to when clicking on the button.
        
        > Mutually exclusive with the `custom_id` and `sku_id` fields.
    
    Returns
    -------
    button : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
        - `url` and `custom_id` are mutually exclusive.
    """
    return Component(
        ComponentType.button,
        button_style = style,
        custom_id = custom_id,
        label = label,
        enabled = enabled,
        emoji = emoji,
        sku_id = sku_id,
        url = url,
    )


def create_row(
    *components,
):
    """
    Creates a row component from the given components.
    
    Parameters
    ----------
    *components : ``Component``
        Sub components.
    
    Returns
    -------
    row : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    """
    return Component(
        ComponentType.row,
        components = components,
    )


def create_string_select(
    options,
    custom_id = ...,
    *,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new string select.
    
    Parameters
    ----------
    options : `None` or `iterable` of ``StringSelectOption``
        Options of the select.
    
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    string_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.string_select,
        custom_id = custom_id,
        enabled = enabled,
        options = options,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,required = required,
    )


def create_text_input(
    label = ...,
    *,
    custom_id = ...,
    max_length = ...,
    min_length = ...,
    placeholder = ...,
    required = ...,
    style = ...,
    value = ...,
):
    """
    Creates a new text input. Can be used inside of forms.
    
    Parameters
    ----------
    label : `None | str`, Optional
        Label of the component.
    
    custom_id : `None | str`, Optional (Keyword only)
        Custom identifier to detect which text input was clicked by the user.
        
        If not given, is autogenerated form the `label` parameter.
    
    max_length : `int`, Optional (Keyword only)
        The maximal length of the inputted text.
    
    min_length : `int`, Optional (Keyword only)
        The minimal length of the inputted text.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the input.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
        
        If not given, or given as `None`, will default to `True` if `min_length` is defined as higher than `0`.
    
    style : ``None | int | TextInputStyle``, Optional (Keyword only)
        The text input's style.
    
    value : `None | str`, Optional (Keyword only)
        The text input's default value.
    
    Returns
    -------
    text_input : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.text_input,
        custom_id = custom_id,
        label = label,
        max_length = max_length,
        min_length = min_length,
        placeholder = placeholder,
        required = required,
        text_input_style = style,
        value = value,
    )


def create_user_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new user select.
    
    Parameters
    ----------
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<ClientUserBase | EntitySelectDefaultValue | tuple>`, Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    user_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.user_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
        required = required,
    )


def create_role_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new role select.
    
    Parameters
    ----------
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<Role | EntitySelectDefaultValue | tuple>`, Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    role_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.role_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
        required = required,
    )


def create_mentionable_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new mentionable (user & role) select.
    
    Parameters
    ----------
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<ClientUserBase | Role | EntitySelectDefaultValue | tuple>` \
            , Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    mentionable_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.mentionable_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
        required = required,
    )


def create_channel_select(
    custom_id = ...,
    *,
    channel_types = ...,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new channel select.
    
    Parameters
    ----------
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the user.
    
    channel_types : `None`, `tuple` of (``ChannelType``, `int`), Optional (Keyword only)
        The allowed channel types by the select.
    
    default_values : `None | iterable<ClientUserBase | Channel | EntitySelectDefaultValue | tuple>` \
            , Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None | str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    channel_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.channel_select,
        channel_types = channel_types,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
        required = required,
    )


def create_section(
    *components,
    thumbnail = ...,
):
    """
    Creates a section component from the given components.
    
    Parameters
    ----------
    *components : ``Component``
        Sub components.
    
    thumbnail : ``None | Component``, Optional (Keyword only)
        The thumbnail or other accessory (button) of a section component.
    
    Returns
    -------
    row : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    """
    return Component(
        ComponentType.section,
        components = components,
        thumbnail = thumbnail,
    )


def create_text_display(
    content = ...,
):
    """
    Creates a new text display component. Can be used inside of forms.
    
    Parameters
    ----------
    content : `None | str`, Optional
        The content shown on the component.
    
    Returns
    -------
    text : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.text_display,
        content = content,
    )


def create_thumbnail_media(
    media,
    *,
    description = ...,
    spoiler = ...,
):
    """
    Creates a new thumbnail media component. Can be used inside of forms.
    
    Parameters
    ----------
    media : ``str | MediaInfo``
        The media of the component.
    
    description : `None | str`, Optional (Keyword only)
        Description of the component's media.
    
    spoiler : `bool` Optional (Keyword only)
        Whether the media should be spoilered.
    
    Returns
    -------
    separator : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.thumbnail_media,
        description = description,
        media = media,
        spoiler = spoiler,
    )


def create_media_gallery(
    *items,
):
    """
    Creates a new media gallery component. Can be used inside of forms.
    
    Parameters
    ----------
    *items : ``str | MediaItem``
        The media items shown on the component.
    
    Returns
    -------
    media_gallery : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.media_gallery,
        items = items,
    )


def create_attachment_media(
    media,
    *,
    spoiler = ...,
):
    """
    Creates a new attachment media component. Can be used inside of forms.
    
    Parameters
    ----------
    media : ``str | MediaInfo``
        The media of the component.
    
    spoiler : `bool` Optional (Keyword only)
        Whether the media should be spoilered.
    
    Returns
    -------
    separator : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.attachment_media,
        media = media,
        spoiler = spoiler,
    )


def create_separator(
    *,
    divider = ...,
    spacing_size = ...,
):
    """
    Creates a new separator component. Can be used inside of forms.
    
    Parameters
    ----------
    divider : `bool`, Optional (Keyword only)
        Whether the separator should contain a divider.
    
    spacing_size : ``int | SeparatorSpacingSize``, Optional (Keyword only)
        The separator's spacing's size.
    
    Returns
    -------
    separator : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.separator,
        divider = divider,
        spacing_size = spacing_size,
    )


def create_container(
    *components,
    color = ...,
    spoiler = ...,
):
    """
    Creates a container component from the given components.
    
    Parameters
    ----------
    *components : ``Component``
        Sub components.
    
    color : ``None | Color``
        The color of the strip on the left.
    
    spoiler : `bool`, Optional (Keyword only)
        Whether the content of the component is spoilered.
    
    Returns
    -------
    container : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.container,
        components = components,
        color = color,
        spoiler = spoiler,
    )


def create_label(
    label = ...,
    description = ...,
    component = ...,
):
    """
    Creates a new label component.
    
    Parameters
    ----------
    label : `None | str`, Optional
        Label of the component.
    
    description : `None | str`, Optional
        Description of the component.
    
    component : ``None | Component``, Optional
        The encapsulated component.
    
    Returns
    -------
    container : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.label,
        label = label,
        description = description,
        component = component,
    )


def create_attachment_input(
    custom_id = ...,
    *,
    max_values = ...,
    min_values = ...,
    required = ...,
):
    """
    Creates a new attachment input.
    
    Parameters
    ----------
    custom_id : `None | str`, Optional
        Custom identifier to detect which component was used by the file.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to upload.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to upload.
    
    required : `None | bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
    
    Returns
    -------
    attachment_input : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.attachment_input,
        custom_id = custom_id,
        max_values = max_values,
        min_values = min_values,
        required = required,
    )



__all__ = ('RoleColorConfiguration',)

from scarletio import RichAttributeErrorBaseType

from ...color import Color

from .fields import (
    parse_color_primary, parse_color_secondary, parse_color_tertiary, put_color_primary, put_color_secondary,
    put_color_tertiary, validate_color_primary, validate_color_secondary, validate_color_tertiary,
)


class RoleColorConfiguration(RichAttributeErrorBaseType):
    """
    Role color configuration.
    
    Attributes
    ----------
    color_primary : ``Color``
        A role's primary color.
    
    color_secondary : ``None | Color``
        A role's secondary color.
    
    color_tertiary : ``None | Color``
        A role's tertiary color.
    """
    __slots__ = ('color_primary', 'color_secondary', 'color_tertiary')
    
    def __new__(cls, *, color_primary = ..., color_secondary = ..., color_tertiary = ...):
        """
        Creates a role color configuration with the given fields.
        
        Parameters
        ----------
        color_primary : ``None | int | Color``
            A role's primary color.
        
        color_secondary : ``None | int | Color``
            A role's secondary color.
        
        color_tertiary : ``None | int | Color``
            A role's tertiary color.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        
        ValueError
            - If a parameter's value is incorrect.
        """
        # color_primary
        if color_primary is ...:
            color_primary = Color()
        else:
            color_primary = validate_color_primary(color_primary)
        
        # color_secondary
        if color_secondary is ...:
            color_secondary = None
        else:
            color_secondary = validate_color_secondary(color_secondary)
        
        # color_tertiary
        if color_tertiary is ...:
            color_tertiary = None
        else:
            color_tertiary = validate_color_tertiary(color_tertiary)
        
        # Construct
        self = object.__new__(cls)
        self.color_primary = color_primary
        self.color_secondary = color_secondary
        self.color_tertiary = color_tertiary
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # color_primary
        color_primary = self.color_primary
        if color_primary:
            repr_parts.append(' color_primary = ')
            repr_parts.append(repr(color_primary))
            
            field_added = True
        else:
            field_added = False

        # color_secondary
        color_secondary = self.color_secondary
        if (color_secondary is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' color_secondary = ')
            repr_parts.append(repr(color_secondary))
            
        
        # color_tertiary
        color_tertiary = self.color_tertiary
        if (color_tertiary is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' color_tertiary = ')
            repr_parts.append(repr(color_tertiary))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        # color_primary
        hash_value = hash(self.color_primary)
        
        # color_secondary
        color_secondary = self.color_secondary
        if (color_secondary is not None):
            hash_value ^= 1
            hash_value ^= hash(color_secondary)

        # color_tertiary
        color_tertiary = self.color_tertiary
        if (color_tertiary is not None):
            hash_value ^= 2
            hash_value ^= hash(color_tertiary)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # color_primary
        if self.color_primary != other.color_primary:
            return False
        
        # color_secondary
        if self.color_secondary != other.color_secondary:
            return False
        
        # color_tertiary
        if self.color_tertiary != other.color_tertiary:
            return False
        
        return True
    
    
    @classmethod
    def create_empty(cls):
        """
        Creates an empty role color configuration.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.color_primary = Color()
        self.color_secondary = None
        self.color_tertiary = None
        return self
    
    
    @classmethod
    def create_from_color_primary(cls, color_primary):
        """
        Creates an empty role color configuration.
        
        Parameters
        ----------
        color_primary : ``Color``
            Primary color to create instance with.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.color_primary = color_primary
        self.color_secondary = None
        self.color_tertiary = None
        return self
    
    
    def to_data(self, defaults = False):
        """
        Converts the role color configuration into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_color_primary(self.color_primary, data, defaults)
        put_color_secondary(self.color_secondary, data, defaults)
        put_color_tertiary(self.color_tertiary, data, defaults)
        return data
    
    
    @classmethod
    def from_data(cls, data):
        """
        Parses a role configuration out from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.color_primary = parse_color_primary(data)
        self.color_secondary = parse_color_secondary(data)
        self.color_tertiary = parse_color_tertiary(data)
        return self
    
    
    def copy(self):
        """
        Copies the role color configuration.
        
        Returns
        -------
        self : `instance<cls>`
        """
        new = object.__new__(type(self))
        new.color_primary = self.color_primary
        new.color_secondary = self.color_secondary
        new.color_tertiary = self.color_tertiary
        return new
    
    
    def copy_with(self, *, color_primary = ..., color_secondary = ..., color_tertiary = ...):
        """
        Copies the role color configuration with the given fields.
        
        Parameters
        ----------
        color_primary : ``None | int | Color``
            A role's primary color.
        
        color_secondary : ``None | int | Color``
            A role's secondary color.
        
        color_tertiary : ``None | int | Color``
            A role's tertiary color.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        
        ValueError
            - If a parameter's value is incorrect.
        """
        # color_primary
        if color_primary is ...:
            color_primary = self.color_primary
        else:
            color_primary = validate_color_primary(color_primary)
        
        # color_secondary
        if color_secondary is ...:
            color_secondary = self.color_secondary
        else:
            color_secondary = validate_color_secondary(color_secondary)
        
        # color_tertiary
        if color_tertiary is ...:
            color_tertiary = self.color_tertiary
        else:
            color_tertiary = validate_color_tertiary(color_tertiary)
        
        # Construct
        new = object.__new__(type(self))
        new.color_primary = color_primary
        new.color_secondary = color_secondary
        new.color_tertiary = color_tertiary
        return new



from .flag import *
from .flag_deprecation import *
from .flag_descriptors import *
from .flag_meta import *


__all__ = (
    *flag.__all__,
    *flag_deprecation.__all__,
    *flag_descriptors.__all__,
    *flag_meta.__all__,
)



__all__ = ('AuditLog', )

from scarletio import RichAttributeErrorBaseType, WeakReferer, copy_func, export, set_docs

from ...core import GUILDS

from .fields import (
    parse_application_commands, parse_auto_moderation_rules, parse_entries, parse_integrations, parse_scheduled_events,
    parse_threads, parse_users, parse_webhooks, put_application_commands, put_auto_moderation_rules,
    put_entries, put_integrations, put_scheduled_events, put_threads, put_users,
    put_webhooks, validate_application_commands, validate_auto_moderation_rules, validate_entries,
    validate_guild_id, validate_integrations, validate_scheduled_events, validate_threads, validate_users,
    validate_webhooks
)
from .helpers import _merge_dictionaries, _merge_lists


@export
class AuditLog(RichAttributeErrorBaseType):
    """
    Whenever an admin action is performed on the API, an audit log entry is added to the respective guild's audit
    logs. This class represents a requested  collections of these entries.
    
    Attributes
    ----------
    _self_reference : `None | WeakReferer<instance>`
        Weak reference to the audit log itself.
    application_commands : `None`, `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the mentioned application commands by the audit log entries. The keys are the `id`-s
        of the application commands, meanwhile the values are the application commands themselves.
    auto_moderation_rules : `None`, `dict` of (`int`, ``ApplicationCommand``) items
        A dictionary that contains the mentioned auto moderation rules by the audit log's entries. The keys
        are the `id`-s of the rules and the values are the rules themselves.
    entries : `None`, `list` of ``AuditLogEntry``
        A list of audit log entries that the audit log contains.
    guild_id : `int`
        The audit logs' respective guild's identifier.
    integrations : `None`, `dict` of (`int`, ``Integration``) items
        A dictionary that contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    scheduled_events : `None`, `dict` of (`int`, ``ScheduledEvent``) items
        A dictionary that containing the mentioned scheduled events by the the audit log's entries.
    threads : `None`, `dict` of (`int`, ``Channel``) items
        A dictionary that containing the mentioned threads by the audit log's entries.
    users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        A dictionary that contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `None`, `dict` of (`int`, ``Webhook``) items
        A dictionary that contains the mentioned webhook by the audit log's entries. The keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    """
    __slots__ = (
        '__weakref__', '_self_reference', 'application_commands', 'auto_moderation_rules', 'entries', 'guild_id',
        'integrations', 'scheduled_events', 'threads', 'users', 'webhooks'
    )
    
    def __new__(
        cls,
        application_commands = ...,
        auto_moderation_rules = ...,
        entries = ...,
        guild_id = ...,
        integrations = ...,
        scheduled_events = ...,
        threads = ...,
        users = ...,
        webhooks = ...,
    ):
        """
        Creates a new audit log instance.
        
        Parameters
        ----------
        application_commands : `None`, `iterable` of ``ApplicationCommand``, Optional (Keyword only)
            The mentioned application commands by the audit log's entries.
        auto_moderation_rules : `None`, `iterable` of ``ApplicationCommand``, Optional (Keyword only)
            The mentioned auto moderation rules inside of the audit log's entries.
        entries : `None`, `iterable` of ``AuditLogEntry``, Optional (Keyword only)
            The audit log entries that the audit log contains.
        guild_id : `int`
            The audit log' respective guild's identifier.
        integrations : `None`, `iterable` of ``Integration``, Optional (Keyword only)
            The mentioned integrations by the audit log's entries
        scheduled_events : `None`, `iterable` of ``ScheduledEvent``, Optional (Keyword only)
            The mentioned scheduled events mentioned by the audit log's entries.
        threads : `None`, `iterable` of ``Channel``, Optional (Keyword only)
            The mentioned threads inside of the audit log's entries.
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The mentioned users by the audit log's entries.
        webhooks : `None`, `iterable` of ``Webhook`, Optional (Keyword only)
            The mentioned webhook by the audit log's entries.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_commands
        if application_commands is ...:
            application_commands = None
        else:
            application_commands = validate_application_commands(application_commands)
        
        # auto_moderation_rules
        if auto_moderation_rules is ...:
            auto_moderation_rules = None
        else:
            auto_moderation_rules = validate_auto_moderation_rules(auto_moderation_rules)
        
        # entries
        if entries is ...:
            entries = None
        else:
            entries = validate_entries(entries)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # integrations
        if integrations is ...:
            integrations = None
        else:
            integrations = validate_integrations(integrations)
        
        # scheduled_events
        if scheduled_events is ...:
            scheduled_events = None
        else:
            scheduled_events = validate_scheduled_events(scheduled_events)
        
        # threads
        if threads is ...:
            threads = None
        else:
            threads = validate_threads(threads)
        
        # users
        if users is ...:
            users = None
        else:
            users = validate_users(users)
        
        # webhooks
        if webhooks is ...:
            webhooks = None
        else:
            webhooks = validate_webhooks(webhooks)
        
        # Construct
        self = object.__new__(cls)
        self._self_reference = None
        self.application_commands = application_commands
        self.auto_moderation_rules = auto_moderation_rules
        self.entries = entries
        self.guild_id = guild_id
        self.integrations = integrations
        self.scheduled_events = scheduled_events
        self.threads = threads
        self.users = users
        self.webhooks = webhooks
        
        # Postprocess
        for entry in self.iter_entries():
            entry._link_parent_soft(self)
        
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates an audit log from the data,
        
        Parameters
        ----------
        data : `dict<str, object>`
            Audit log data.
        guild_id : `int` = `0`, Optional
            The respective guild's identifier of the audit logs.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._self_reference = None
        self.application_commands = parse_application_commands(data)
        self.auto_moderation_rules = parse_auto_moderation_rules(data)
        self.entries = None
        self.guild_id = guild_id
        self.integrations = parse_integrations(data)
        self.scheduled_events = parse_scheduled_events(data)
        self.threads = parse_threads(data, guild_id)
        self.users = parse_users(data)
        self.webhooks = parse_webhooks(data)
        
        # We need self for entries.
        self.entries = parse_entries(data, self)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serialises the audit log.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_application_commands(self.application_commands, data, defaults)
        put_auto_moderation_rules(self.auto_moderation_rules, data, defaults)
        put_entries(self.entries, data, defaults)
        put_integrations(self.integrations, data, defaults)
        put_scheduled_events(self.scheduled_events, data, defaults)
        put_threads(self.threads, data, defaults)
        put_users(self.users, data, defaults)
        put_webhooks(self.webhooks, data, defaults)
        return data
    
    
    @classmethod
    def from_many(cls, audit_logs):
        """
        Creates a new audit log from many audit logs.
        
        Parameters
        ----------
        audit_logs : `list<AuditLog>`
            Audit logs to merge.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._self_reference = None
        self.application_commands = _merge_dictionaries(audit_log.application_commands for audit_log in audit_logs)
        self.auto_moderation_rules = _merge_dictionaries(audit_log.auto_moderation_rules for audit_log in audit_logs)
        self.entries = _merge_lists(audit_log.entries for audit_log in audit_logs)
        self.guild_id = audit_logs[0].guild_id if audit_logs else 0
        self.integrations = _merge_dictionaries(audit_log.integrations for audit_log in audit_logs)
        self.scheduled_events = _merge_dictionaries(audit_log.scheduled_events for audit_log in audit_logs)
        self.threads = _merge_dictionaries(audit_log.threads for audit_log in audit_logs)
        self.users = _merge_dictionaries(audit_log.users for audit_log in audit_logs)
        self.webhooks = _merge_dictionaries(audit_log.webhooks for audit_log in audit_logs)
        return self
    
    
    def copy(self):
        """
        Copies the audit log.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._self_reference = None
        
        # application_commands
        application_commands = self.application_commands
        if (application_commands is not None):
            application_commands = application_commands.copy()
        new.application_commands = application_commands
        
        # auto_moderation_rules
        auto_moderation_rules = self.auto_moderation_rules
        if (auto_moderation_rules is not None):
            auto_moderation_rules = auto_moderation_rules.copy()
        new.auto_moderation_rules = auto_moderation_rules
        
        # entries
        entries = self.entries
        if (entries is not None):
            entries = [entry.copy() for entry in entries]
        new.entries = entries
        
        # guild_id
        new.guild_id = self.guild_id
        
        # integrations
        integrations = self.integrations
        if (integrations is not None):
            integrations = integrations.copy()
        new.integrations = integrations
        
        # scheduled_events
        scheduled_events = self.scheduled_events
        if (scheduled_events is not None):
            scheduled_events = scheduled_events.copy()
        new.scheduled_events = scheduled_events
        
        # threads
        threads = self.threads
        if (threads is not None):
            threads = threads.copy()
        new.threads = threads
        
        # users
        users = self.users
        if (users is not None):
            users = users.copy()
        new.users = users
        
        # webhooks
        webhooks = self.webhooks
        if (webhooks is not None):
            webhooks = webhooks.copy()
        new.webhooks = webhooks
        
        # Postprocess
        for entry in new.iter_entries():
            entry._link_parent_hard(new)
        
        return new
    
    
    def copy_with(
        self,
        application_commands = ...,
        auto_moderation_rules = ...,
        entries = ...,
        guild_id = ...,
        integrations = ...,
        scheduled_events = ...,
        threads = ...,
        users = ...,
        webhooks = ...,
    ):
        """
        Copies the audit log with the given fields.
        
        Parameters
        ----------
        application_commands : `None`, `iterable` of ``ApplicationCommand``, Optional (Keyword only)
            The mentioned application commands by the audit log's entries.
        auto_moderation_rules : `None`, `iterable` of ``ApplicationCommand``, Optional (Keyword only)
            The mentioned auto moderation rules inside of the audit log's entries.
        entries : `None`, `iterable` of ``AuditLogEntry``, Optional (Keyword only)
            The audit log entries that the audit log contains.
        guild_id : `int`
            The audit log' respective guild's identifier.
        integrations : `None`, `iterable` of ``Integration``, Optional (Keyword only)
            The mentioned integrations by the audit log's entries
        scheduled_events : `None`, `iterable` of ``ScheduledEvent``, Optional (Keyword only)
            The mentioned scheduled events mentioned by the audit log's entries.
        threads : `None`, `iterable` of ``Channel``, Optional (Keyword only)
            The mentioned threads inside of the audit log's entries.
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The mentioned users by the audit log's entries.
        webhooks : `None`, `iterable` of ``Webhook`, Optional (Keyword only)
            The mentioned webhook by the audit log's entries.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_commands
        if application_commands is ...:
            application_commands = self.application_commands
            if (application_commands is not None):
                application_commands = application_commands.copy()
        else:
            application_commands = validate_application_commands(application_commands)
        
        # auto_moderation_rules
        if auto_moderation_rules is ...:
            auto_moderation_rules = self.auto_moderation_rules
            if (auto_moderation_rules is not None):
                auto_moderation_rules = auto_moderation_rules.copy()
        else:
            auto_moderation_rules = validate_auto_moderation_rules(auto_moderation_rules)
        
        # entries
        if entries is ...:
            entries = self.entries
            if (entries is not None):
                entries = [entry.copy() for entry in entries]
            
            entries_new = True
        else:
            entries = validate_entries(entries)
            entries_new = False
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # integrations
        if integrations is ...:
            integrations = self.integrations
            if (integrations is not None):
                integrations = integrations.copy()
        else:
            integrations = validate_integrations(integrations)
        
        # scheduled_events
        if scheduled_events is ...:
            scheduled_events = self.scheduled_events
            if (scheduled_events is not None):
                scheduled_events = scheduled_events.copy()
        else:
            scheduled_events = validate_scheduled_events(scheduled_events)
        
        # threads
        if threads is ...:
            threads = self.threads
            if (threads is not None):
                threads = threads.copy()
        else:
            threads = validate_threads(threads)
        
        # users
        if users is ...:
            users = self.users
            if (users is not None):
                users = users.copy()
        else:
            users = validate_users(users)
        
        # webhooks
        if webhooks is ...:
            webhooks = self.webhooks
            if (webhooks is not None):
                webhooks = webhooks.copy()
        else:
            webhooks = validate_webhooks(webhooks)
        
        # Construct
        new = object.__new__(type(self))
        new._self_reference = None
        new.application_commands = application_commands
        new.auto_moderation_rules = auto_moderation_rules
        new.entries = entries
        new.guild_id = guild_id
        new.integrations = integrations
        new.scheduled_events = scheduled_events
        new.threads = threads
        new.users = users
        new.webhooks = webhooks
        
        # Postprocess
        if entries_new:
            for entry in new.iter_entries():
                entry._link_parent_hard(new)
        else:
            for entry in new.iter_entries():
                entry._link_parent_soft(new)
        
        return new
    
    
    def __iter__(self):
        """Iterates over the audit log's entries."""
        entries = self.entries
        if (entries is not None):
            yield from entries
    
    
    def __reversed__(self):
        """Reversed iterator over the audit log's entries."""
        entries = self.entries
        if (entries is not None):
            yield from reversed(entries)
    
    
    def __len__(self):
        """Returns the amount of entries that the audit log contain."""
        entries = self.entries
        if (entries is None):
            return 0
        
        return len(entries)
    
    
    def __getitem__(self, index):
        """Returns the specific audit log entry at the given index."""
        entries = self.entries
        if entries is None:
            if isinstance(index, int):
                raise IndexError(index)
            
            if isinstance(index, slice):
                return []
            
            raise TypeError(
                f'`index` can be either `int` or `slice`, got {type(index).__name__}; {index!r}.'
            )
        
        return entries[index]
    
    
    def __repr__(self):
        """Returns the representation of the Audit log."""
        return f'<{type(self).__name__} length = {len(self)}>'
    
    
    def __bool__(self):
        """Returns whether the audit log entry contains at least 1 element."""
        return (self.entries is not None)
    
    
    def __eq__(self, other):
        """Returns whether the two audit logs are equal."""
        # Shortcut
        if self is other:
            return True
        
        # Check type
        if type(self) is not type(other):
            return NotImplemented
        
        # First check guild_id and entries since they are the one who probably diverge.
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # entries
        if self.entries != other.entries:
            return False
        
        # application_commands
        if self.application_commands != other.application_commands:
            return False
        
        # auto_moderation_rules
        if self.auto_moderation_rules != other.auto_moderation_rules:
            return False
        
        # integrations
        if self.integrations != other.integrations:
            return False
        
        # scheduled_events
        if self.scheduled_events != other.scheduled_events:
            return False
        
        # threads
        if self.threads != other.threads:
            return False
        
        # users
        if self.users != other.users:
            return False
        
        # webhooks
        if self.webhooks != other.webhooks:
            return False
        
        return True
    
    
    def _get_self_reference(self):
        """
        Returns a self (weak) reference to the audit log itself.
        
        Returns
        -------
        self_reference : ``WeakReferer``
        """
        self_reference = self._self_reference
        if (self_reference is None):
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
        
        return self_reference
    
    
    @property
    def guild(self):
        """
        Returns the audit log's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        return GUILDS.get(self.guild_id, None)
    
    
    # Getters
    
    def get_application_command(self, application_command_id):
        """
        Gets the application command for the given identifier.
        
        Parameters
        ----------
        application_command_id : `int`
            Application command identifier.
        
        Returns
        -------
        application_command : `None | ApplicationCommand`
        """
        application_commands = self.application_commands
        if (application_commands is not None):
            return application_commands.get(application_command_id, None)
    
    
    def get_auto_moderation_rule(self, auto_moderation_rule_id):
        """
        Gets the auto moderation rule for the given identifier.
        
        Parameters
        ----------
        auto_moderation_rule_id : `int`
            Auto moderation rule identifier.
        
        Returns
        -------
        auto_moderation_rule : `None | AutoModerationRule`
        """
        auto_moderation_rules = self.auto_moderation_rules
        if (auto_moderation_rules is not None):
            return auto_moderation_rules.get(auto_moderation_rule_id, None)
    
    
    def get_integration(self, integration_id):
        """
        Gets the integration for the given identifier.
        
        Parameters
        ----------
        integration_id : `int`
            integration identifier.
        
        Returns
        -------
        integration : `None | Integration`
        """
        integrations = self.integrations
        if (integrations is not None):
            return integrations.get(integration_id, None)
    
    
    def get_scheduled_event(self, scheduled_event_id):
        """
        Gets the scheduled event for the given identifier.
        
        Parameters
        ----------
        scheduled_event_id : `int`
            Scheduled event identifier.
        
        Returns
        -------
        scheduled_event : `None | ScheduledEvent`
        """
        scheduled_events = self.scheduled_events
        if (scheduled_events is not None):
            return scheduled_events.get(scheduled_event_id, None)
    
    
    def get_thread(self, thread_id):
        """
        Gets the thread for the given identifier.
        
        Parameters
        ----------
        thread_id : `int`
            Thread identifier.
        
        Returns
        -------
        thread : `None | Channel`
        """
        threads = self.threads
        if (threads is not None):
            return threads.get(thread_id, None)
    
    
    def get_user(self, user_id):
        """
        Gets the user for the given identifier.
        
        Parameters
        ----------
        user_id : `int`
            User identifier.
        
        Returns
        -------
        user : `None | ClientUserBase`
        """
        users = self.users
        if (users is not None):
            return users.get(user_id, None)
    
    
    def get_webhook(self, Webhook_id):
        """
        Gets the webhook for the given identifier.
        
        Parameters
        ----------
        webhook_id : `int`
            Webhook identifier.
        
        Returns
        -------
        webhook : `None | Webhook`
        """
        webhooks = self.webhooks
        if (webhooks is not None):
            return webhooks.get(Webhook_id, None)
    
    
    # Iterators
    
    
    def iter_application_commands(self):
        """
        Iterates over the application commands.
        
        This method is an iterable generator.
        
        Yields
        ------
        application_command : ``ApplicationCommand``
        """
        application_commands = self.application_commands
        if (application_commands is not None):
            yield from application_commands.values()
    
    
    def iter_auto_moderation_rules(self):
        """
        Iterates over the auto moderation rules.
        
        This method is an iterable generator.
        
        Yields
        ------
        auto_moderation_rule : ``AutoModerationRule``
        """
        auto_moderation_rules = self.auto_moderation_rules
        if (auto_moderation_rules is not None):
            yield from auto_moderation_rules.values()
    
    
    iter_entries = set_docs(
        copy_func(__iter__),
        """
        Iterates over the entries.
        
        This method is an iterable generator.
        
        Yields
        ------
        entry : ``AuditLogEntry``
        """
    )
    
    
    def iter_integrations(self):
        """
        Iterates over the integrations.
        
        This method is an iterable generator.
        
        Yields
        ------
        integration : ``Integration``
        """
        integrations = self.integrations
        if (integrations is not None):
            yield from integrations.values()
    
    
    def iter_scheduled_events(self):
        """
        Iterates over the scheduled events.
        
        This method is an iterable generator.
        
        Yields
        ------
        scheduled_event : ``ScheduledEvent``
        """
        scheduled_events = self.scheduled_events
        if (scheduled_events is not None):
            yield from scheduled_events.values()    
    
    
    def iter_threads(self):
        """
        Iterates over the threads.
        
        This method is an iterable generator.
        
        Yields
        ------
        thread : ``Channel``
        """
        threads = self.threads
        if (threads is not None):
            yield from threads.values()
    
    
    def iter_users(self):
        """
        Iterates over the users.
        
        This method is an iterable generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        users = self.users
        if (users is not None):
            yield from users.values()
    
    
    def iter_webhooks(self):
        """
        Iterates over the webhooks.
        
        This method is an iterable generator.
        
        Yields
        ------
        webhook : ``Webhook``
        """
        webhooks = self.webhooks
        if (webhooks is not None):
            yield from webhooks.values()



__all__ = ('AutoModerationRuleTriggerMetadataMentionSpam',)

from scarletio import copy_docs

from .base import AutoModerationRuleTriggerMetadataBase
from .constants import AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
from .fields import (
    parse_mention_limit, parse_raid_protection, put_mention_limit, put_raid_protection,
    validate_mention_limit, validate_raid_protection
)


class AutoModerationRuleTriggerMetadataMentionSpam(AutoModerationRuleTriggerMetadataBase):
    """
    Mention spam trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    mention_limit : `int`
        The amount of mentions in a message after the rule is triggered.
    raid_protection : `bool`
        Whether mention raid protection is enabled.
    """
    __slots__ = ('mention_limit', 'raid_protection')
    
    def __new__(cls, mention_limit = ..., *, raid_protection = ...):
        """
        Creates a new mention spam trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        mention_limit : `int`, Optional
            The amount of mentions in a message after the rule is triggered.
            
            Defaults to the allowed maximal amount.
        raid_protection : `bool`, Optional
            Whether mention raid protection is enabled.
        
        Raises
        ------
        TypeError
            - If `mention_limit` type is incorrect.
        """
        # mention_limit
        if mention_limit is ...:
            mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
        else:
            mention_limit = validate_mention_limit(mention_limit)
        
        # raid_protection
        if raid_protection is ...:
            raid_protection = False
        else:
            raid_protection = validate_raid_protection(raid_protection)
        
        # Construct
        self = object.__new__(cls)
        self.mention_limit = mention_limit
        self.raid_protection = raid_protection
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # mention_limit
        repr_parts.append(' mention_limit = ')
        repr_parts.append(repr(self.mention_limit))
        
        raid_protection = self.raid_protection
        if raid_protection:
            repr_parts.append(', raid_protection = ')
            repr_parts.append(repr(raid_protection))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.mention_limit = parse_mention_limit(data)
        self.raid_protection = parse_raid_protection(data)
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_mention_limit(self.mention_limit, data, defaults)
        put_raid_protection(self.raid_protection, data, defaults)
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # mention_limit
        if self.mention_limit != other.mention_limit:
            return False
        
        # raid_protection
        if self.raid_protection != other.raid_protection:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # mention_limit
        hash_value ^= self.mention_limit
        
        # raid_protection
        hash_value ^= self.raid_protection << 6
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.mention_limit = self.mention_limit
        new.raid_protection = self.raid_protection
        return new
    
    
    def copy_with(self, *, mention_limit = ..., raid_protection = ...):
        """
        Copies the trigger metadata with altering it's attributes based on the given fields.
        
        Parameters
        ----------
        mention_limit : `int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
        raid_protection : `bool`, Optional
            Whether mention raid protection is enabled.
        
        Returns
        -------
        new : `instance<type<self>>`
               
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        """
        # mention_limit
        if mention_limit is ...:
            mention_limit = self.mention_limit
        else:
            mention_limit = validate_mention_limit(mention_limit)
        
        # raid_protection
        if raid_protection is ...:
            raid_protection = self.raid_protection
        else:
            raid_protection = validate_raid_protection(raid_protection)
        
        new = object.__new__(type(self))
        new.mention_limit = mention_limit
        new.raid_protection = raid_protection
        return new



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, int_postprocess_parser_factory, nullable_date_time_parser_factory,
    nullable_entity_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, int_optional_postprocess_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_entity_array_putter_factory, nullable_entity_putter_factory, preinstanced_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, int_conditional_validator_factory, nullable_date_time_validator_factory,
    preinstanced_validator_factory
)

from ..poll_answer import PollAnswer
from ..poll_question import PollQuestion
from ..poll_result import PollResult
from ..poll_result.utils import merge_update_poll_results

from .constants import DURATION_DEFAULT, DURATION_MAX
from .preinstanced import PollLayout


# allow_multiple_choices

parse_allow_multiple_choices = bool_parser_factory('allow_multiselect', False)
put_allow_multiple_choices = bool_optional_putter_factory('allow_multiselect', False)
validate_allow_multiple_choices = bool_validator_factory('allow_multiple_choices', False)


# answers

def parse_answers(data):
    """
    Parses answers from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    answers : `None | tuple<PollAnswer>`
    """
    answer_datas = data.get('answers', None)
    if (answer_datas is None) or (not answer_datas):
        return None
    
    return (*(PollAnswer.from_data(answer_data) for answer_data in answer_datas),)    


put_answers = nullable_entity_array_putter_factory('answers', PollAnswer)


def validate_answers(answers):
    """
    Validates the given answers.
    
    Parameters
    ----------
    answers : `None | iterable<PollAnswer | str>`
        The answers to validate.
    
    Returns
    -------
    answers : `None | tuple<PollAnswer>`
    """
    if answers is None:
        return None
    
    if (getattr(answers, '__iter__', None) is None):
        raise TypeError(
            f'`answers` can be `None`, `iterable` of `{PollAnswer.__name__}`, `str`, got '
            f'{type(answers).__name__}; {answers!r}.'
        )
        
    answers_validated = None
    
    for answer in answers:
        if isinstance(answer, PollAnswer):
            pass
        
        elif isinstance(answer, str):
            answer = PollAnswer(text = answer)
        
        else:
            raise TypeError(
                f'`answers` can contain `{PollAnswer.__name__}`, `str` elements, got '
                f'{type(answer).__name__}; {answer!r}; answers = {answers!r}.'
            )
        
        if (answers_validated is None):
            answers_validated = []
        
        answers_validated.append(answer)
    
    if (answers_validated is not None):
        return tuple(answers_validated)

# duration

parse_duration = int_postprocess_parser_factory(
    'duration',
    DURATION_DEFAULT,
    (lambda duration: duration * 3600),
)
put_duration = int_optional_postprocess_putter_factory(
    'duration',
    DURATION_DEFAULT,
    (lambda duration: duration // 3600),
)
validate_duration = int_conditional_validator_factory(
    'duration',
    DURATION_DEFAULT,
    (lambda duration: duration > 0 and duration <= DURATION_MAX and duration % 3600 == 0),
    f'must be > 0 and <= and {DURATION_MAX} multiple of 3600'
)


# expires_at

parse_expires_at = nullable_date_time_parser_factory('expiry')
put_expires_at = nullable_date_time_optional_putter_factory('expiry')
validate_expires_at = nullable_date_time_validator_factory('expires_at')

# finalized


def parse_finalized(data):
    """
    Parses out the finalized value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    finalized : `bool`
    """
    nested_data = data.get('results', None)
    if nested_data is None:
        return False
    
    finalized = nested_data.get('is_finalized', None)
    if finalized is None:
        return False
    
    return finalized


def put_finalized(finalized, data, defaults):
    """
    Serializes the finalized value into the given `data`.
    
    Parameters
    ----------
    finalized : `bool`
        Value to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    nested_data = data.get('results', None)
    if nested_data is None:
        nested_data = {}
        data['results'] = nested_data
    
    nested_data['is_finalized'] = finalized
    
    return data


validate_finalized = bool_validator_factory('finalized', False)

# layout

parse_layout = preinstanced_parser_factory('layout_type', PollLayout, PollLayout.default)
put_layout = preinstanced_optional_putter_factory('layout_type', PollLayout.default)
validate_layout = preinstanced_validator_factory('layout', PollLayout)


# question

parse_question = nullable_entity_parser_factory('question', PollQuestion)
put_question = nullable_entity_putter_factory('question', PollQuestion)


def validate_question(question):
    """
    Validates whether the given `question` is valid.
    
    Parameters
    ----------
    question : `None`, ``PollQuestion``, `str`
        The question to validate.
    
    Returns
    -------
    question : `None`, ``PollQuestion``
        The validated question.
    
    Raises
    ------
    TypeError
    ValueError
    """
    if question is None:
        return None
    
    if isinstance(question, PollQuestion):
        return question
    
    if isinstance(question, str):
        return PollQuestion(text = question)
    
    raise TypeError(
        f'`question` can be `None`, `{PollQuestion.__name__}`, `str`, got {type(question).__name__!s}; {question!r}.'
    )


# results

def parse_results(data, old_results = None):
    """
    Parses the poll results from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    old_results : `None | list<PollResult>` = `None`, Optional
        Old results of the poll.
    
    Returns
    -------
    results : `None | list<PollResult>`
    """
    nested_data = data.get('results', None)
    if nested_data is None:
        return None
    
    result_datas = nested_data.get('answer_counts', None)
    if (result_datas is None) or (not result_datas):
        new_results = None
    else:
        new_results = [PollResult.from_data(result_data) for result_data in result_datas]
    return merge_update_poll_results(new_results, old_results)


def put_results(results, data, defaults):
    """
    Serializes poll results into the given `data`.
    
    Parameters
    ----------
    results : `None | list<PollResult>`
        Results to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if results is None:
        result_datas = []
    else:
        result_datas = [result.to_data(defaults = defaults) for result in results]
    
    nested_data = data.get('results', None)
    if nested_data is None:
        nested_data = {}
        data['results'] = nested_data
    
    nested_data['answer_counts'] = result_datas
    
    return data


def validate_results(results):
    """
    Validates the given results.
    
    Parameters
    ----------
    results : `None | iterable<PollResult>`
        The results to validate.
    
    Returns
    -------
    results : `None | list<PollResult>`
    """
    if results is None:
        return None
    
    if (getattr(results, '__iter__', None) is None):
        raise TypeError(
            f'`results` can be `None`, `iterable` of `{PollResult.__name__}`, got '
            f'{type(results).__name__}; {results!r}.'
        )
        
    results_validated = None
    
    for result in results:
        if not isinstance(result, PollResult):
            raise TypeError(
                f'`results` can contain `{PollResult.__name__}` elements, got '
                f'{type(result).__name__}; {result!r}; results = {results!r}.'
            )
        
        if (results_validated is None):
            results_validated = []
        
        results_validated.append(result)
    
    return results_validated



__all__ = ()

from types import CodeType


VALUABLE_CODE_FIELDS = (CodeType.co_consts, CodeType.co_code, CodeType.co_names)


def _hash_function(function):
    """
    Hashes the give function.
    
    Parameters
    ----------
    function : `None | FunctionType | MethodType`
        The function to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    if function is None:
        return 0
    
    code = function.__code__
    
    hash_value = 0
    
    for descriptor in VALUABLE_CODE_FIELDS:
        hash_value ^= hash(type(descriptor).__get__(descriptor, code))
    
    return hash_value


def _eq_functions(function_0, function_1):
    """
    Returns whether the two given functions are the same.
    
    Parameters
    ----------
    function_0 : `None | FunctionType | MethodType`
        Function to equal with.
    function_1 : `None | FunctionType | MethodType`
        Function to equal with.
    
    Returns
    -------
    are_equal : `bool`
    """
    if function_0 is function_1:
        return True
    
    if (function_0 is None) or (function_1 is None):
        return False
    
    code_0 = function_0.__code__
    code_1 = function_1.__code__
    
    for descriptor in VALUABLE_CODE_FIELDS:
        getter = type(descriptor).__get__
        if getter(descriptor, code_0) != getter(descriptor, code_1):
            return False
    
    return True


SORT_KEY = lambda item: item[0]


def _hash_dict(value):
    """
    Hashes the given dictionary.
    
    Parameters
    ----------
    value : `dict<hashable & sortable, hashable>`
        Value to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    return  hash((*sorted(value.items(), key = SORT_KEY),))


def _hash_change_value(value):
    """
    Hashes an audit log change's `.before` or `.after` value.
    
    Parameters
    ----------
    value : `None | hashable | dict<hashable & sortable, hashable>`
        Value to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    if value is None:
        hash_value = 0
    elif isinstance(value, dict):
        hash_value = _hash_dict(value)
    else:
        hash_value = hash(value)

    return hash_value



__all__ = ()

from vampytest import ScarletioCoroutineEnvironment, set_global_environment

from .discord import KOKORO


# We run every async test inside of hata on our event loop.
set_global_environment(ScarletioCoroutineEnvironment(event_loop = KOKORO))



__all__ = ('CONVERSION_STICKER',)

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion
from ....sticker import Sticker, create_partial_sticker_from_id

from .sticker_ids import CONVERSION_STICKER_IDS


class CONVERSION_STICKER(Conversion):
    # Generic
    
    name = 'sticker'
    name_aliases = None
    expected_types_messages = (
        f'`None`, `{Sticker.__name__}`, `int`, `(list | tuple)<{Sticker.__name__}>`, `(list | tuple)<int>`'
    )
    kind = CONVERSION_KIND_FIELD
    output_conversion = CONVERSION_STICKER_IDS
    
    # Setting
    
    set_merger = None
    set_type = Sticker
    
    def set_type_processor(value):
        return [value.id]
    
    
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield None
            return
        
        # Sticker
        if isinstance(value, Sticker):
            yield [value.id]
        
        # int
        if isinstance(value, int):
            yield [value]
            return
        
        # (list | tuple)<int | Sticker>
        if isinstance(value, list) or isinstance(value, tuple):
            sticker_ids = None
            
            for element in value:
                if isinstance(element, Sticker):
                    element = element.id
                
                elif isinstance(element, int):
                    pass
                
                else:
                    return
            
                if sticker_ids is None:
                    sticker_ids = []
                    
                sticker_ids.append(element)
                continue
            
            yield sticker_ids
            return
        
        # No other cases
        return
    
    
    # Reading
    
    get_default = None
    
    def get_processor(value):
        if value is not None:
            return create_partial_sticker_from_id(value[0])
    
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    
    # Sorting
    
    sort_priority = 1600



__all__ = ('ComponentMetadataMediaGallery',)

from scarletio import copy_docs

from .base import ComponentMetadataBase
from .fields import parse_items, put_items, validate_items


class ComponentMetadataMediaGallery(ComponentMetadataBase):
    """
    Represents the metadata of a media gallery component that is mainly used to show multiple images.
    
    Attributes
    ----------
    items : ``None | tuple<MediaItem>``
        The media items shown on the component.
    """
    __slots__ = ('items',)
    
    def __new__(cls, *, items = ...):
        """
        Creates a media gallery component metadata.
        
        Parameters
        ----------
        items : ``None | iterable<str> | iterable<MediaItem>``, Optional (Keyword only)
            The media items shown on the component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # items
        if items is ...:
            items = None
        else:
            items = validate_items(items)
        
        # Construct
        self = object.__new__(cls)
        self.items = items
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            items = keyword_parameters.pop('items', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # items
        items = self.items
        if (items is not None):
            repr_parts.append(' items = ')
            repr_parts.append(repr(items))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # items
        items = self.items
        if (items is not None):
            hash_value ^= len(items)
            
            for item in items:
                hash_value ^= hash(item)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # items
        if self.items != other.items:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.items = parse_items(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_items(self.items, data, defaults, include_internals = include_internals)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        
        items = self.items
        if (items is not None):
            items = (*(item.copy() for item in self.items),)
        new.items = items
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        items = self.items
        if (items is not None):
            items = (*(item.copy() for item in self.items),)
        new.items = items
        
        return new
    
    
    def copy_with(
        self, 
        *,
        items = ...,
    ):
        """
        Copies the media gallery component metadata with the given fields.
        
        Parameters
        ----------
        items : ``None | iterable<str> | iterable<MediaItem>``, Optional (Keyword only)
            The media items shown on the component.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # items
        if items is ...:
            items = self.items
            if (items is not None):
                items = (*(item.copy() for item in self.items),)
        else:
            items = validate_items(items)
        
        # Construct
        new = object.__new__(type(self))
        new.items = items
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            items = keyword_parameters.pop('items', ...),
        )



from .embedded_activity import *
from .embedded_activity_location import *
from .embedded_activity_user_state import *


__all__ = (
    *embedded_activity.__all__,
    *embedded_activity_location.__all__,
    *embedded_activity_user_state.__all__,
)



__all__ = ('ActivityParty',)

from scarletio import copy_docs

from ..activity_field_base import ActivityFieldBase

from .fields import (
    parse_id, parse_size_and_max, put_id, put_size_and_max, validate_id, validate_max, validate_size
)


class ActivityParty(ActivityFieldBase):
    """
    Represents a discord activity party.
    
    Attributes
    ----------
    id : `None`, `str`
        The party's id, which in the player is.
    size : `int`
        The party's maximal size in which the player is. Defaults to `0`.
    max : `int`
        The party's actual size in which the player is. Defaults to `0`.
    """
    __slots__ = ('id', 'size', 'max',)
    
    def __new__(cls, *, party_id = ..., size = ..., max_ = ...):
        """
        Creates a new activity party instance form the given parameters.
        
        Parameters
        ----------
        party_id : `None`, `str`, Optional (Keyword only)
            The party's id, which in the player is.
        size : `int`, Optional (Keyword only)
            The party's maximal size, which in the player is.
        max_ : `int`, Optional (Keyword only)
            The party's actual size, which in the player is.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # party_id
        if party_id is ...:
            party_id = None
        else:
            party_id = validate_id(party_id)
        
        # size
        if size is ...:
            size = 0
        else:
            size = validate_size(size)
        
        # max_
        if max_ is ...:
            max_ = 0
        else:
            max_ = validate_max(max_)
        
        # Construct
        self = object.__new__(cls)
        self.id = party_id
        self.size = size
        self.max = max_
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.size, self.max = parse_size_and_max(data)
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_id(self.id, data, defaults)
        put_size_and_max((self.size, self.max), data, defaults)
        return data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        party_id = self.id
        if (party_id is not None):
            repr_parts.append(' id = ')
            repr_parts.append(repr(party_id))
            field_added = True
        else:
            field_added = False
        
        size = self.size
        max_ = self.max
        if size or max_:
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' size = ')
            repr_parts.append(repr(size))
            repr_parts.append(', max = ')
            repr_parts.append(repr(max_))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.size != other.size:
            return False
        
        if self.max != other.max:
            return False
        
        return True
    

    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        party_id = self.id
        if (party_id is not None):
            hash_value ^= hash(party_id)
            hash_value ^= (1 << 0)
        
        size = self.size
        if size:
            hash_value ^= size << 4
            hash_value ^= (1 << 1)
        
        max_ = self.max
        if max_:
            hash_value ^= max_ << 8
            hash_value ^= (1 << 2)
        
        return hash_value
    
    
    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        if (self.id is not None):
            return True
        
        size = self.size
        if size:
            return True
        
        max_ = self.max
        if max_:
            return True
        
        return False
    
    
    @copy_docs(ActivityFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.id = self.id
        new.size = self.size
        new.max = self.max
        return new
    
    
    def copy_with(self, party_id = ..., id_ = ..., size = ..., max_ = ...):
        """
        Copies the activity party with the given fields.
        
        Parameters
        ----------
        party_id : `None`, `str`, Optional (Keyword only)
            The party's id, which in the player is.
        size : `int`, Optional (Keyword only)
            The party's maximal size, which in the player is.
        max_ : `int`, Optional (Keyword only)
            The party's actual size, which in the player is.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """

        # party_id
        if party_id is ...:
            party_id = self.id
        else:
            party_id = validate_id(party_id)
        
        # size
        if size is ...:
            size = self.size
        else:
            size = validate_size(size)
        
        # max_
        if max_ is ...:
            max_ = self.max
        else:
            max_ = validate_max(max_)
        
        # Construct
        new = object.__new__(type(self))
        new.id = party_id
        new.size = size
        new.max = max_
        return new



__all__ = ('InteractionComponent',)

from scarletio import RichAttributeErrorBaseType, copy_docs, export, include

from ..component import ComponentType
from ..interaction_component_metadata import InteractionComponentMetadataBase
from ..interaction_component_metadata.fields import (
    validate_component__label, validate_components, validate_custom_id, validate_thumbnail, validate_value,
    validate_values
)

from .fields import parse_type, put_type, validate_type


Component = include('Component')


@export
class InteractionComponent(RichAttributeErrorBaseType):
    """
    Integration component representing a parameter or a sub-command-group.
    
    Attributes
    ----------
    type : ``ComponentType``
        The represented component's type.
    
    metadata : ``InteractionComponentMetadataBase``
        The interaction component's metadata.
    """
    __slots__ = ('metadata', 'type')
    
    
    def __new__(cls, component_type, **keyword_parameters):
        """
        Creates a new interaction component from the given keyword parameters.
        
        Parameters
        ----------
        component_type : ``None | int | ComponentType``
            The component's type to create.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining the component's fields.
        
        Other Parameters
        ----------------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The sub-component nested inside (the label component).
        
        components : ``None | tuple<InteractionComponent>``, Optional (Keyword only)
            Sub-components.
        
        custom_id : `None | str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        thumbnail : ``None | InteractionComponent``, Optional (Keyword only)
            The thumbnail or other accessory (button) of a section component.
        
        value : `None | str`, Optional (Keyword only)
            The component's value defined by the user.
        
        values : `None | str`, Optional (Keyword only)
            The component's values selected by the user.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a field's value is incorrect.
        """
        component_type = validate_type(component_type)
        metadata = component_type.interaction_metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        self = object.__new__(cls)
        self.type = component_type
        self.metadata = metadata
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an interaction component with it's default values set.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.type = ComponentType.none
        self.metadata = InteractionComponentMetadataBase()
        return self
    
    
    def copy(self):
        """
        Copies the interaction component.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        new.metadata = self.metadata.copy()
        new.type = self.type
        return new
    
    
    def copy_with(self, *, component_type = ..., **keyword_parameters):
        """
        Copies the interaction component with replacing the defined fields.
        
        Parameters
        ----------
        component_type : ``None | int | ComponentType``, Optional (Keyword only)
            The component's type to create.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining the component's fields.
        
        Other Parameters
        ----------------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The sub-component nested inside (the label component).
        
        components : ``None | tuple<InteractionComponent>``, Optional (Keyword only)
            Sub-components.
        
        custom_id : `None | str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        thumbnail : ``None | InteractionComponent``, Optional (Keyword only)
            The thumbnail or other accessory (button) of a section component.
        
        value : `None | str`, Optional (Keyword only)
            The component's value defined by the user.
        
        values : `None | str`, Optional (Keyword only)
            The component's values selected by the user.
        
        Returns
        -------
        new : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = self.type
        else:
            component_type = validate_type(component_type)
        
        # metadata
        if component_type is self.type:
            metadata = self.metadata.copy_with_keyword_parameters(keyword_parameters)
        else:
            metadata = component_type.interaction_metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new.metadata = metadata
        new.type = component_type
        return new
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction component from the data received from Discord.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Interaction component data.
        """
        component_type = parse_type(data)
        metadata = component_type.interaction_metadata_type.from_data(data)
        
        self = object.__new__(cls)
        self.metadata = metadata
        self.type = component_type
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the interaction component into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        # metadata
        data = self.metadata.to_data(defaults = defaults)
        
        # type
        put_type(self.type, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' type = ')
        
        component_type = self.type
        repr_parts.append(component_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(component_type.value))
        
        repr_parts.append(', metadata = ')
        repr_parts.append(repr(self.metadata))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # metadata
        hash_value ^= hash(self.metadata)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self.metadata != other.metadata:
            return False
        
        return True
    
    
    def __mod__(self, other):
        """Returns self % other."""
        if type(other) is not Component:
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        return self.metadata._match_to_component(other)
    
    
    __rmod__ = __mod__
    
    
    def iter_components(self):
        """
        Iterates over the sub-components of the interaction component.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``InteractionComponent``
        """
        components = self.components
        if (components is not None):
            yield from components


    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the interaction component.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id_is_multi_value_values : `(str, ComponentType, None | str | tuple<str>)`
        """
        return (yield from self.metadata.iter_custom_ids_and_values())
    
    
    # component
    @property
    @copy_docs(InteractionComponentMetadataBase.component)
    def component(self):
        return self.metadata.component
    
    @component.setter
    def component(self, component):
        self.metadata.component = validate_component__label(component)
    
    
    # components
    @property
    @copy_docs(InteractionComponentMetadataBase.components)
    def components(self):
        return self.metadata.components
    
    @components.setter
    def components(self, components):
        self.metadata.components = validate_components(components)
    
    
    # custom_id
    @property
    @copy_docs(InteractionComponentMetadataBase.custom_id)
    def custom_id(self):
        return self.metadata.custom_id
    
    @custom_id.setter
    def custom_id(self, custom_id):
        self.metadata.custom_id = validate_custom_id(custom_id)
    
    
    # thumbnail
    @property
    @copy_docs(InteractionComponentMetadataBase.thumbnail)
    def thumbnail(self):
        return self.metadata.thumbnail
    
    @thumbnail.setter
    def thumbnail(self, thumbnail):
        self.metadata.thumbnail = validate_thumbnail(thumbnail)

    
    # value
    @property
    @copy_docs(InteractionComponentMetadataBase.value)
    def value(self):
        return self.metadata.value
    
    @value.setter
    def value(self, value):
        self.metadata.value = validate_value(value)
    
    
    # values
    @property
    @copy_docs(InteractionComponentMetadataBase.values)
    def values(self):
        return self.metadata.values
    
    @values.setter
    def values(self, values):
        self.metadata.values = validate_values(values)



__all__ = ()

from ...channel import Channel
from ...field_parsers import entity_id_parser_factory, preinstanced_parser_factory
from ...field_putters import entity_id_optional_putter_factory, preinstanced_putter_factory
from ...field_validators import entity_id_validator_factory, preinstanced_validator_factory

from .preinstanced import EmbeddedActivityLocationType


# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_optional_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')


# type

parse_type = preinstanced_parser_factory('kind', EmbeddedActivityLocationType, EmbeddedActivityLocationType.none)
put_type = preinstanced_putter_factory('kind')
validate_type = preinstanced_validator_factory('type', EmbeddedActivityLocationType)



__all__ = ()

from ...field_parsers import entity_id_parser_factory
from ...field_putters import entity_id_optional_putter_factory, entity_id_putter_factory
from ...field_validators import entity_id_validator_factory, entity_validator_factory

from ..scheduled_event import ScheduledEvent
from ..scheduled_event_occasion_overwrite import ScheduledEventOccasionOverwrite


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')


# scheduled_event_id

parse_scheduled_event_id = entity_id_parser_factory('event_id')
put_scheduled_event_id = entity_id_putter_factory('event_id')
validate_scheduled_event_id = entity_id_validator_factory('scheduled_event_id', ScheduledEvent)


# occasion_overwrite

def parse_occasion_overwrite(data):
    """
    Parses an occasion overwrite out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    occasion_overwrite : ``ScheduledEventOccasionOverwrite``
    """
    return ScheduledEventOccasionOverwrite.from_data(data)


def put_occasion_overwrite(occasion_overwrite, data, defaults):
    """
    Serializes the given scheduled event occasion overwrite.
    
    Parameters
    ----------
    occasion_overwrite : ``ScheduledEventOccasionOverwrite``
        Occasion overwrite to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data.update(occasion_overwrite.to_data(defaults = defaults, include_internals = True))
    return data


validate_occasion_overwrite = entity_validator_factory('occasion_overwrite', ScheduledEventOccasionOverwrite)



__all__ = ('ComponentMetadataEntitySelectBase', )

from scarletio import copy_docs

from .fields import parse_default_values, put_default_values, validate_default_values
from .select_base import ComponentMetadataSelectBase


class ComponentMetadataEntitySelectBase(ComponentMetadataSelectBase):
    """
    Channel select component metadata.
    
    Attributes
    ----------
    custom_id : `None | str`
        Custom identifier to detect which component was used by the user.
    
    default_values : ``None | tuple<EntitySelectDefaultValue>``
        Entities presented in the select by default.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    placeholder : `None | str`
        Placeholder text of the select.
    
    required : `bool`
        Whether the field is required to be fulfilled.
    """
    __slots__ = ('default_values',)
    
    
    def __new__(
        cls,
        *,
        custom_id = ...,
        default_values = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        placeholder = ...,
        required = ...,
    ):
        """
        Creates a new channel select component metadata with the given parameters.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        default_values : ``None | iterable<Channel> | iterable<Role> | iterable<ClientUserBase> | iterable<EntitySelectDefaultValue> | iterable<(str | EntitySelectDefaultValueTyp, int | str)>>`` \
                , Optional (Keyword only)
            Entities presented in the select by default.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None | str`, Optional (Keyword only)
            Placeholder text of the select.
        
        required : `None | bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_values
        if default_values is ...:
            default_values = None
        else:
            default_values = validate_default_values(default_values)
        
        # Construct
        self = ComponentMetadataSelectBase.__new__(
            cls,
            custom_id = custom_id,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
            required = required,
        )
        self.default_values = default_values
        
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataSelectBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            custom_id = keyword_parameters.pop('custom_id', ...),
            default_values = keyword_parameters.pop('default_values', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @copy_docs(ComponentMetadataSelectBase._add_type_specific_repr_fields_into)
    def _add_type_specific_repr_fields_into(self, repr_parts):
        # default_values
        repr_parts.append(', default_values = ')
        default_values = self.default_values
        if (default_values is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(default_values)
            
            while True:
                option = default_values[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
    
    
    @copy_docs(ComponentMetadataSelectBase.__hash__)
    def __hash__(self):
        hash_value = ComponentMetadataSelectBase.__hash__(self)
        
        # default_values
        default_values = self.default_values
        if (default_values is not None):
            hash_value ^= hash(default_values)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataSelectBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ComponentMetadataSelectBase._is_equal_same_type(self, other):
            return False
        
        # default_values
        if self.default_values != other.default_values:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataSelectBase.from_data)
    def from_data(cls, data):
        self = super(ComponentMetadataEntitySelectBase, cls).from_data(data)
        self.default_values = parse_default_values(data)
        return self
    
    
    @copy_docs(ComponentMetadataSelectBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ComponentMetadataSelectBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        put_default_values(self.default_values, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataSelectBase.clean_copy)
    def clean_copy(self, guild = None):
        new = ComponentMetadataSelectBase.clean_copy(self, guild)
        
        # default_values
        default_values = self.default_values
        if (default_values is not None):
            default_values = (*default_values,)
        new.default_values = default_values
        
        return new
    
    
    @copy_docs(ComponentMetadataSelectBase.copy)
    def copy(self):
        new = ComponentMetadataSelectBase.copy(self)
        
        # default_values
        default_values = self.default_values
        if (default_values is not None):
            default_values = (*default_values,)
        new.default_values = default_values
        
        return new
    
    
    def copy_with(
        self,
        *,
        custom_id = ...,
        default_values = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        placeholder = ...,
        required = ...,
    ):
        """
        Copies the channel select component metadata with the given fields.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        default_values : ``None | iterable<Channel> | iterable<Role> | iterable<ClientUserBase> | iterable<EntitySelectDefaultValue> | iterable<(str | EntitySelectDefaultValueTyp, int | str)>>`` \
                , Optional (Keyword only)
            Entities presented in the select by default.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None | str`, Optional (Keyword only)
            Placeholder text of the select.
        
        required : `None | bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_values
        if default_values is ...:
            default_values = self.default_values
            if (default_values is not None):
                default_values = (*default_values,)
        else:
            default_values = validate_default_values(default_values)
        
        # Construct
        
        new = ComponentMetadataSelectBase.copy_with(
            self,
            custom_id = custom_id,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
            required = required,
        )
        new.default_values = default_values
        return new
    
    
    @copy_docs(ComponentMetadataSelectBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            custom_id = keyword_parameters.pop('custom_id', ...),
            default_values = keyword_parameters.pop('default_values', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
            required = keyword_parameters.pop('required', ...),
        )



__all__ = ('AutoModerationRule',)

from ...bases import DiscordEntity
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import AUTO_MODERATION_RULES, GUILDS
from ...role import create_partial_role_from_id
from ...precreate_helpers import process_precreate_parameters
from ...user import ZEROUSER, create_partial_user_from_id

from ..trigger_metadata import AutoModerationRuleTriggerMetadataBase

from .fields import (
    parse_actions, parse_creator_id, parse_enabled, parse_event_type, parse_excluded_channel_ids,
    parse_excluded_role_ids, parse_guild_id, parse_id, parse_name, parse_trigger_metadata, parse_trigger_type,
    put_actions, put_creator_id, put_enabled, put_event_type, put_excluded_channel_ids,
    put_excluded_role_ids, put_guild_id, put_id, put_name, put_trigger_metadata,
    put_trigger_type, validate_actions, validate_creator_id, validate_enabled, validate_event_type,
    validate_excluded_channel_ids, validate_excluded_role_ids, validate_guild_id, validate_id, validate_name,
    validate_trigger_type
)
from .helpers import guess_rule_trigger_type_from_keyword_parameters
from .preinstanced import AutoModerationEventType, AutoModerationRuleTriggerType


PRECREATE_FIELDS = {
    'actions': ('actions', validate_actions),
    'creator_id': ('creator_id', validate_creator_id),
    'enabled': ('enabled', validate_enabled),
    'event_type': ('event_type', validate_event_type),
    'excluded_channel_ids': ('excluded_channel_ids', validate_excluded_channel_ids),
    'excluded_role_ids': ('excluded_role_ids', validate_excluded_role_ids),
    'guild_id': ('guild_id', validate_guild_id),
    'name': ('name', validate_name),
}


class AutoModerationRule(DiscordEntity, immortal = True):
    """
    Auto moderation feature which allows guilds to set rules that trigger based on some criteria.
    
    Attributes
    ----------
    id : `int`
        The unique identifier of the auto moderation rule. Defaults to `0`.
    
    actions : `None`, `tuple` of ``AutoModerationAction``
        Actions which will execute when the rule is triggered. Defaults to `0`.
    
    creator_id : `int`
        The user who created the rule. Defaults to `0`.
    
    enabled : `bool`
        Whether the rule is enabled.
    
    event_type : ``AutoModerationEventType``
        For which events is the rule applied.
    
    excluded_channel_ids : `None | tuple<int>`
        The excluded channels' identifiers.
    
    excluded_role_ids : `None | tuple<int>`
        The excluded roles' identifiers.
    
    guild_id : `int`
        The guild's identifier where the rule is. Defaults to `0`.
    
    name : `str`
        The rule's name.
    
    trigger_metadata : ``AutoModerationRuleTriggerMetadataBase``
        Trigger type specific metadata if applicable.
    
    trigger_type : ``AutoModerationRuleTriggerType``
        Characterizes the type of content which can trigger the rule.
    """
    __slots__ = (
        'actions', 'creator_id', 'enabled', 'event_type', 'excluded_channel_ids', 'excluded_role_ids', 'guild_id',
        'name', 'trigger_metadata', 'trigger_type'
    )
    
    
    def __new__(
        cls,
        name = ...,
        actions = ...,
        trigger_type = ...,
        *,
        enabled = ...,
        event_type = ...,
        excluded_channel_ids = ...,
        excluded_role_ids = ...,
        **keyword_parameters,
    ):
        """
        Creates a new auto moderation rule with the given parameters.
        
        Parameters
        ----------
        name : `str`, Optional
            The rule's name.
        
        actions : `None`, ``AutoModerationAction``, `iterable` of ``AutoModerationAction``, Optional
             Actions which will execute when the rule is triggered.
        
        trigger_type : ``AutoModerationRuleTriggerType``, `int`, Optional
            Auto moderation trigger type.
        
        enabled : `bool` , Optional (Keyword only)
            Whether the rule is enabled.
        
        event_type : ``None | AutoModerationEventType | int``, Optional (Keyword only)
            For which events is the rule applied.
        
        excluded_channel_ids : ``None | int | Channel | iterable<int> | iterable<Channel>``, Optional (Keyword only)
            Excluded channels from the rule.
        
        excluded_role_ids : ``None | int | Role | iterable<int> | iterable<Role>``, Optional (Keyword only)
            Excluded roles from the rule.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters defining how the rule is triggered.
        
        Other Parameters
        ----------------
        excluded_keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Excluded keywords from preset filter.
        
        keyword_presets : ``None | int, AutoModerationKeywordPresetType | iterable<int> | iterable<AutoModerationKeywordPresetType>`` \
                , Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
            
        keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Substrings which will be searched for in content.
        
        mention_limit : `None | int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
        
        regex_patterns : `None | tuple<str>`, Optional (Keyword only)
            Regular expression patterns which are matched against content.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # actions
        if actions is ...:
            actions = None
        else:
            actions = validate_actions(actions)
        
        # enabled
        if enabled is ...:
            enabled = True
        else:
            enabled = validate_enabled(enabled)
        
        # event_type
        if event_type is ...:
            event_type = AutoModerationEventType.message_send
        else:
            event_type = validate_event_type(event_type)
        
        # excluded_channel_ids
        if excluded_channel_ids is ...:
            excluded_channel_ids = None
        else:
            excluded_channel_ids = validate_excluded_channel_ids(excluded_channel_ids)
        
        # excluded_role_ids
        if excluded_role_ids is ...:
            excluded_role_ids = None
        else:
            excluded_role_ids = validate_excluded_role_ids(excluded_role_ids)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # trigger_type
        if trigger_type is ...:
            trigger_type = AutoModerationRuleTriggerType.none
        else:
            trigger_type = validate_trigger_type(trigger_type)
        
        # **keyword_parameters
        trigger_type = guess_rule_trigger_type_from_keyword_parameters(trigger_type, keyword_parameters)
        trigger_metadata = trigger_type.metadata_type(**keyword_parameters)
        
        self = object.__new__(cls)
        self.actions = actions
        self.creator_id = 0
        self.enabled = enabled
        self.event_type = event_type
        self.excluded_channel_ids = excluded_channel_ids
        self.excluded_role_ids = excluded_role_ids
        self.guild_id = 0
        self.id = 0
        self.name = name
        self.trigger_metadata = trigger_metadata
        self.trigger_type = trigger_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``AutoModerationRule`` from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received auto moderation rule data.
        
        Returns
        -------
        self : ``AutoModerationRule``
            The created auto moderation rule.
        """
        auto_moderation_rule_id = parse_id(data)
        
        try:
            self = AUTO_MODERATION_RULES[auto_moderation_rule_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = auto_moderation_rule_id
            self._set_attributes(data)
            AUTO_MODERATION_RULES[auto_moderation_rule_id] = self
        
        else:
            self._update_attributes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the auto moderation rule to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
                
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_actions(self.actions, data, defaults)
        
        if include_internals:
            put_creator_id(self.creator_id, data, defaults)
        
        put_enabled(self.enabled, data, defaults)
        put_event_type(self.event_type, data, defaults)
        put_excluded_channel_ids(self.excluded_channel_ids, data, defaults)
        put_excluded_role_ids(self.excluded_role_ids, data, defaults)
        
        if include_internals:
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
        
        put_name(self.name, data, defaults)
        put_trigger_metadata(self.trigger_metadata, data, defaults)
        put_trigger_type(self.trigger_type, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the auto moderation rule's attributes (excluding ``.id``).
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild profile data.
        """        
        self.creator_id = parse_creator_id(data)
        self.guild_id = parse_guild_id(data)
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the auto moderation rule with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild profile data.
        """
        self.actions = parse_actions(data)
        # creator_id | Internal field & cannot change
        self.enabled = parse_enabled(data)
        self.event_type = parse_event_type(data)
        self.excluded_channel_ids = parse_excluded_channel_ids(data)
        self.excluded_role_ids = parse_excluded_role_ids(data)
        # guild_id | Internal field & cannot change
        self.name = parse_name(data)
        # id | Already set & cannot change
        self.trigger_type = trigger_type = parse_trigger_type(data)
        self.trigger_metadata = parse_trigger_metadata(data, trigger_type)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the auto moderation rule and returns the changed attributes in a dictionary with the changed attributes
        in a `attribute-name`, `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild profile data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
            
            Might contain the following items:
            
            +-------------------------------+-----------------------------------------------------------+
            | Keys                          | Values                                                    |
            +===============================+===========================================================+
            | actions                       | `None`, `tuple` of ``AutoModerationAction``               |
            +-------------------------------+-----------------------------------------------------------+
            | enabled                       | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | event_type                    | ``AutoModerationEventType``                               |
            +-------------------------------+-----------------------------------------------------------+
            | excluded_channel_ids          | `None | tuple<int>`                                       |
            +-------------------------------+-----------------------------------------------------------+
            | excluded_role_ids             | `None | tuple<int>`                                       |
            +-------------------------------+-----------------------------------------------------------+
            | name                          | `str`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | trigger_metadata              | ``AutoModerationRuleTriggerMetadataBase``                 |
            +-------------------------------+-----------------------------------------------------------+
            | trigger_type                  | ``AutoModerationRuleTriggerType``                         |
            +-------------------------------+-----------------------------------------------------------+
        """
        old_attributes = {}
        
        # actions
        actions = parse_actions(data)
        if self.actions != actions:
            old_attributes['actions'] = self.actions
            self.actions = actions
        
        # creator_id |Internal field & cannot change
        
        # enabled
        enabled = parse_enabled(data)
        if self.enabled != enabled:
            old_attributes['enabled'] = self.enabled
            self.enabled = enabled
        
        # event_type
        event_type = parse_event_type(data)
        if self.event_type != event_type:
            old_attributes['event_type'] = self.event_type
            self.event_type = event_type
        
        # excluded_channel_ids
        excluded_channel_ids = parse_excluded_channel_ids(data)
        if self.excluded_channel_ids != excluded_channel_ids:
            old_attributes['excluded_channel_ids'] = self.excluded_channel_ids
            self.excluded_channel_ids = excluded_channel_ids
        
        # excluded_role_ids
        excluded_role_ids = parse_excluded_role_ids(data)
        if self.excluded_role_ids != excluded_role_ids:
            old_attributes['excluded_role_ids'] = self.excluded_role_ids
            self.excluded_role_ids = excluded_role_ids
        
        # guild_id | Internal field & cannot change
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # id | Already set & cannot change
        
        # trigger_type
        trigger_type = parse_trigger_type(data)
        if self.trigger_type is not trigger_type:
            old_attributes['trigger_type'] = self.trigger_type
            self.trigger_type = trigger_type
        
        # trigger_metadata
        trigger_metadata = parse_trigger_metadata(data, trigger_type)
        if self.trigger_metadata != trigger_metadata:
            old_attributes['trigger_metadata'] = self.trigger_metadata
            self.trigger_metadata = trigger_metadata
        
        return old_attributes
    
    
    def __eq__(self, other):
        """Returns whether the two auto moderation rules are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id == other_id
        
        # actions
        if self.actions != other.actions:
            return False
        
        # creator_id
        # Internal field
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # event_type
        if self.event_type is not other.event_type:
            return False
        
        # excluded_channel_ids
        if self.excluded_channel_ids != other.excluded_channel_ids:
            return False
        
        # excluded_role_ids
        if self.excluded_role_ids != other.excluded_role_ids:
            return False
        
        # guild_id
        # Internal field
        
        # name
        if self.name != other.name:
            return False
        
        # trigger_metadata
        if self.trigger_metadata != other.trigger_metadata:
            return False
        
        # trigger_type
        if self.trigger_type is not other.trigger_type:
            return False
        
        return True
    
    
    def __ne__(self, other):
        """Returns whether the two auto moderation rules are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id != other_id
        
        # actions
        if self.actions != other.actions:
            return True
        
        # creator_id
        # Internal field
        
        # enabled
        if self.enabled != other.enabled:
            return True
        
        # event_type
        if self.event_type is not other.event_type:
            return True
        
        # excluded_channel_ids
        if self.excluded_channel_ids != other.excluded_channel_ids:
            return True
        
        # excluded_role_ids
        if self.excluded_role_ids != other.excluded_role_ids:
            return True
        
        # guild_id
        # Internal field
        
        # name
        if self.name != other.name:
            return True
        
        # trigger_metadata
        if self.trigger_metadata != other.trigger_metadata:
            return True
        
        # trigger_type
        if self.trigger_type is not other.trigger_type:
            return True
        
        return False
    
    
    def __hash__(self):
        """Returns the auto moderation rule's hash value."""
        
        # id
        self_id = self.id
        if self_id:
            return self_id
        
        hash_value = 0
        
        # actions
        actions = self.actions
        if (actions is not None):
            hash_value ^= len(actions)
            
            for action in actions:
                hash_value ^= hash(action)
        
        # creator_id | Internal field
        
        # enabled
        hash_value ^= self.enabled << 4
        
        # event_type
        hash_value ^= self.event_type.value << 8
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            hash_value ^= len(excluded_channel_ids) << 12
            
            for excluded_channel_id in excluded_channel_ids:
                hash_value ^= excluded_channel_id
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            hash_value ^= len(excluded_role_ids) << 16
            
            for excluded_role_id in excluded_role_ids:
                hash_value ^= excluded_role_id
        
        # guild_id | Internal field
        
        # name
        hash_value ^= hash(self.name)
                
        # trigger_metadata1
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is not None):
            hash_value ^= hash(trigger_metadata)
        
        # trigger_type
        hash_value ^= self.trigger_type.value << 20
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the auto moderation rule's representation"""
        repr_parts = ['<', self.__class__.__name__]
        
        # system fields: `.id`, `.guild_id`, `.creator_id`
        
        # id
        rule_id = self.id
        if rule_id == 0:
            repr_parts.append(' partial')
        
        else:
            repr_parts.append(' id = ')
            repr_parts.append(repr(rule_id))
            
            repr_parts.append(', guild_id = ')
            repr_parts.append(repr(self.guild_id))
            
            repr_parts.append(', creator_id = ')
            repr_parts.append(repr(self.creator_id))
        
        # Information `.name`, `.trigger_type`
        
        # name
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        # trigger_type
        trigger_type = self.trigger_type
        repr_parts.append(', trigger_type = ')
        repr_parts.append(repr(trigger_type.name))
        repr_parts.append(' ~ ')
        repr_parts.append(repr(trigger_type.value))
        
        # Additional: `.actions`, `.enabled`, `event_type`, `.excluded_channel_ids`, `.excluded_role_ids`, 
        # `.trigger_metadata`
        
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is not None):
            repr_parts.append(', trigger_metadata = ')
            repr_parts.append(repr(trigger_metadata))
        
        repr_parts.append(', actions = [')
        actions = self.actions
        if (actions is not None):
            limit = len(actions)
            index = 0
            
            while True:
                action = actions[index]
                repr_parts.append(repr(action))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # enabled
        enabled = self.enabled
        if not enabled:
            repr_parts.append(', enabled = ')
            repr_parts.append(repr(enabled))
        
        event_type = self.event_type
        if (
            (event_type is not AutoModerationEventType.none) and
            (event_type is not AutoModerationEventType.message_send)
        ):
            repr_parts.append(', event_type = ')
            repr_parts.append(repr(event_type.name))
            repr_parts.append('~')
            repr_parts.append(repr(event_type.value))
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            repr_parts.append(', excluded_channel_ids = [')
            
            limit = len(excluded_channel_ids)
            index = 0
            
            while True:
                excluded_channel_id = excluded_channel_ids[index]
                repr_parts.append(repr(excluded_channel_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            repr_parts.append(', excluded_role_ids = [')
            
            limit = len(excluded_role_ids)
            index = 0
            
            while True:
                excluded_role_id = excluded_role_ids[index]
                repr_parts.append(repr(excluded_role_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the auto moderation rule returning a new partial one.
        
        Returns
        -------
        new : ``AutoModerationRule``
        """
        new = object.__new__(type(self))
        
        # id | Internal field
        new.id = 0
        
        # actions
        actions = self.actions
        if (actions is not None):
            actions = tuple(action.copy() for action in actions)
        new.actions = actions
        
        # creator_id | Internal field
        new.creator_id = 0
        
        # enabled
        new.enabled = self.enabled
        
        # event_type
        new.event_type = self.event_type
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            excluded_channel_ids = (*excluded_channel_ids,)
        new.excluded_channel_ids = excluded_channel_ids
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            excluded_role_ids = (*excluded_role_ids,)
        new.excluded_role_ids = excluded_role_ids
        
        # guild_id
        # Internal field
        new.guild_id = 0
        
        # name
        new.name = self.name
        
        # trigger_metadata
        new.trigger_metadata = self.trigger_metadata.copy()
        
        # trigger_type
        new.trigger_type = self.trigger_type
        
        return new
    
    
    def copy_with(
        self,
        *,
        actions = ...,
        enabled = ...,
        event_type = ...,
        excluded_channel_ids = ...,
        excluded_role_ids = ...,
        name = ...,
        trigger_type = ...,
        **keyword_parameters,
    ):
        """
        Copies the auto moderation rule with the given attributes replaced.

        Parameters
        ----------
        actions : `None`, ``AutoModerationAction``, `iterable` of ``AutoModerationAction``, Optional (keyword only)
             Actions which will execute when the rule is triggered.
         
        enabled : `bool` , Optional (Keyword only)
            Whether the rule is enabled.
        
        event_type : ``None | AutoModerationEventType | int``, Optional (Keyword only)
            For which events is the rule applied.
        
        excluded_channel_ids : ``None | int | Channel | iterable<int> | iterable<Channel>``, Optional (Keyword only)
            Excluded channels from the rule.
        
        excluded_role_ids : ``None | int | Role | iterable<int> | iterable<Role>``, Optional (Keyword only)
            Excluded roles from the rule.
        
        name : `str`, Optional (Keyword only)
            The rule's name.
        
        trigger_type : ``AutoModerationRuleTriggerType``, `int`, Optional (Keyword only)
            Auto moderation trigger type.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters defining how the rule is triggered.
        
        Other Parameters
        ----------------
        excluded_keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Excluded keywords from preset filter.
        
        keyword_presets : ``None | int, AutoModerationKeywordPresetType | iterable<int> | iterable<AutoModerationKeywordPresetType>`` \
                , Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
            
        keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Substrings which will be searched for in content.
        
        mention_limit : `None | int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
        
        regex_patterns : `None | tuple<str>`, Optional (Keyword only)
            Regular expression patterns which are matched against content.
        
        Returns
        -------
        new : ``AutoModerationRule``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # actions
        if actions is ...:
            actions = self.actions
            if (actions is not None):
                actions = tuple(action.copy() for action in actions)
        else:
            actions = validate_actions(actions)
        
        # enabled
        if enabled is ...:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # event_type
        if event_type is ...:
            event_type = self.event_type
        else:
            event_type = validate_event_type(event_type)
        
        # excluded_channel_ids
        if excluded_channel_ids is ...:
            excluded_channel_ids = self.excluded_channel_ids
            if (excluded_channel_ids is not None):
                excluded_channel_ids = (*excluded_channel_ids,)
        else:
            excluded_channel_ids = validate_excluded_channel_ids(excluded_channel_ids)
        
        # excluded_role_ids
        if excluded_role_ids is ...:
            excluded_role_ids = self.excluded_role_ids
            if (excluded_role_ids is not None):
                excluded_role_ids = (*excluded_role_ids,)
        else:
            excluded_role_ids = validate_excluded_role_ids(excluded_role_ids)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # trigger_type
        if trigger_type is ...:
            if keyword_parameters:
                trigger_type = guess_rule_trigger_type_from_keyword_parameters(
                    AutoModerationRuleTriggerType.none, keyword_parameters)
            else:
                trigger_type = self.trigger_type
        else:
            trigger_type = validate_trigger_type(trigger_type)
        
        # **keyword_parameters
        if trigger_type is self.trigger_type:
            trigger_metadata = self.trigger_metadata.copy_with(**keyword_parameters)
        else:
            trigger_metadata = trigger_type.metadata_type(**keyword_parameters)
        
        new = object.__new__(type(self))
        
        new.id = 0
        new.actions = actions
        new.creator_id = 0
        new.enabled = enabled
        new.event_type = event_type
        new.excluded_channel_ids = excluded_channel_ids
        new.excluded_role_ids = excluded_role_ids
        new.guild_id = 0
        new.name = name
        new.trigger_metadata = trigger_metadata
        new.trigger_type = trigger_type
        
        return new
    
    
    @classmethod
    def _create_empty(cls, rule_id):
        """
        Creates an auto moderation rule with it's attributes set as their default values.
        
        Parameters
        ----------
        rule_id : `int`
            The auto moderation rule's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = rule_id
        self.actions = None
        self.creator_id = 0
        self.enabled = True
        self.event_type = AutoModerationEventType.none
        self.excluded_channel_ids = None
        self.excluded_role_ids = None
        self.guild_id = 0
        self.name = ''
        self.trigger_metadata = AutoModerationRuleTriggerMetadataBase()
        self.trigger_type = AutoModerationRuleTriggerType.none
        return self
    
    
    @classmethod
    def precreate(cls, rule_id, **keyword_parameters):
        """
        Precreates the auto moderation rule by creating a new one if not yet exists of the given identifier.
        
        Parameters
        ----------
        rule_id : `int`
            The auto moderation rule's identifier.
        
        *keyword_parameters : Keyword parameters
            Keyword parameters defining how should the rule's attributes be set.
        
        Other Parameters
        ----------------
        actions : `None`, ``AutoModerationAction``, `iterable` of ``AutoModerationAction``, Optional (keyword only)
             Actions which will execute when the rule is triggered.
    
        creator_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user who created the rule.
            
        enabled : `bool` , Optional (Keyword only)
            Whether the rule is enabled.
        
        event_type : ``None | AutoModerationEventType | int``, Optional (Keyword only)
            For which events is the rule applied.
        
        excluded_channel_ids : ``None | int | Channel | iterable<int> | iterable<Channel>``, Optional (Keyword only)
            Excluded channels from the rule.
        
        excluded_role_ids : ``None | int | Role | iterable<int> | iterable<Role>``, Optional (Keyword only)
            Excluded roles from the rule.
        
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier where the rule is.
        
        name : `str`, Optional (Keyword only)
            The rule's name.
        
        trigger_type : ``AutoModerationRuleTriggerType``, `int`, Optional (Keyword only)
            Auto moderation trigger type.
        
        excluded_keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Excluded keywords from preset filter.
        
        keyword_presets : ``None | int, AutoModerationKeywordPresetType | iterable<int> | iterable<AutoModerationKeywordPresetType>`` \
                , Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
            
        keywords : `None | str | iterable<str>`, Optional (Keyword only)
            Substrings which will be searched for in content.
        
        mention_limit : `None | int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
        
        regex_patterns : `None | tuple<str>`, Optional (Keyword only)
            Regular expression patterns which are matched against content.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters given.
        ValueError
            - If a parameter's value is incorrect.
        """
        rule_id = validate_id(rule_id)
        
        if keyword_parameters:
            processed = []
            
            # trigger_type
            try:
                trigger_type = keyword_parameters.pop('trigger_type')
            except KeyError:
                trigger_type = AutoModerationRuleTriggerType.none
            else:
                trigger_type = validate_trigger_type(trigger_type)
            
            extra = process_precreate_parameters(keyword_parameters, PRECREATE_FIELDS, processed)
            
            # trigger_metadata
            if (extra is None):
                trigger_metadata = trigger_type.metadata_type()
            else:
                trigger_type = guess_rule_trigger_type_from_keyword_parameters(trigger_type, extra)
                trigger_metadata = trigger_type.metadata_type(**extra)
            
            if (trigger_type is not AutoModerationRuleTriggerType.none):
                processed.append(('trigger_type', trigger_type))
                processed.append(('trigger_metadata', trigger_metadata))
        
        else:
            processed = None
        
        try:
            self = AUTO_MODERATION_RULES[rule_id]
        except KeyError:
            self = cls._create_empty(rule_id)
            AUTO_MODERATION_RULES[rule_id] = self
        
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @property
    def partial(self):
        """
        Returns whether the auto moderation rule is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (not self.id)
    
    
    def iter_actions(self):
        """
        Iterates over the actions of the auto moderation rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        action : ``AutoModerationAction``
        """
        actions = self.actions
        if (actions is not None):
            yield from actions
    
    
    @property
    def creator(self):
        """
        Returns who created the auto moderation rule.
        
        Returns
        -------
        creator : ``ClientUserBase``
        """
        creator_id = self.creator_id
        if creator_id:
            creator = create_partial_user_from_id(creator_id)
        else:
            creator = ZEROUSER
        
        return creator
    
    
    @property
    def guild(self):
        """
        Returns the auto moderation rule's guild. If the guild is not cached returns `None`.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_excluded_channel_ids(self):
        """
        Iterates over the excluded channel identifiers from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_channel_id : `int`
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            yield from excluded_channel_ids
    
    
    @property
    def excluded_channels(self):
        """
        Returns the excluded channels from the rule.
        
        Returns
        -------
        excluded_channels : ``None | tuple<Channel>``
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is None):
            excluded_channels = None
        
        else:
            guild_id = self.guild_id
            excluded_channels = tuple(
                create_partial_channel_from_id(excluded_channel_id, ChannelType.unknown, guild_id)
                for excluded_channel_id in excluded_channel_ids
            )
        
        return excluded_channels
    
    
    def iter_excluded_channels(self):
        """
        Iterates over the excluded channels from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_channel : ``Channel``
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            guild_id = self.guild_id
            
            for excluded_channel_id in excluded_channel_ids:
                yield create_partial_channel_from_id(excluded_channel_id, ChannelType.unknown, guild_id)
    
    
    def iter_excluded_role_ids(self):
        """
        Iterates over the excluded role identifiers from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_role_id : `int`
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            yield from excluded_role_ids
    
    
    @property
    def excluded_roles(self):
        """
        Returns the excluded roles from the rule.
        
        Returns
        -------
        excluded_roles : ``None | tuple<Role>``
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is None):
            excluded_roles = None
        
        else:
            excluded_roles = tuple(
                create_partial_role_from_id(excluded_role_id) for excluded_role_id in excluded_role_ids
            )
        
        return excluded_roles
    
    
    def iter_excluded_roles(self):
        """
        Iterates over the excluded roles from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_role : ``Role``
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            for excluded_role_id in excluded_role_ids:
                yield create_partial_role_from_id(excluded_role_id)



__all__ = ()

from .form_field_keyword import ParameterConverterFormFieldKeyword


class ParameterConverterFormFieldMulti(ParameterConverterFormFieldKeyword):
    """
    Regex and string matcher and `custom_id` matching multi parameter parser for forms.
    
    Attributes
    ----------
    annotation : `str | re.Pattern`
        Annotation defaulting to the parameter's name if required.
    
    default : `object`
        Default value of the parameter.
    
    parameter_name : `str`
        The parameter's name.
    
    matcher : `FunctionType`ParameterConverterFormFieldMulti
        Matches interaction options based on their `custom_id`.
    """
    __slots__ = ()
    
    @staticmethod
    def _converter_string(converter, interaction_event):
        """
        String form submit interaction option multi value matcher.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        values : `None | list<object>`
            The matched values.
        """
        component_type, value_or_values = interaction_event.get_value_for(converter.annotation)
        
        iter_resolve = component_type.iter_resolve
        if (iter_resolve is None):
            values = None
        
        else:
            values = [*iter_resolve(interaction_event.resolved, value_or_values)]
            if not values:
                values = None
        
        return values
    
    
    @staticmethod
    def _converter_regex(converter, interaction_event):
        """
        Regex form submit interaction option multi value matcher.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        values : `None`, `list<object>`
            The matched values.
        """
        values = []
        
        for match, component_type, value_or_values in interaction_event.iter_matches_and_values(
            converter.annotation.fullmatch
        ):
            iter_resolve = component_type.iter_resolve
            if (iter_resolve is None):
                continue
            
            values.extend(iter_resolve(interaction_event.resolved, value_or_values))
            continue
        
        if not values:
            values = None
        
        return values
    
    
    @staticmethod
    def _converter_regex_group_dict(converter, interaction_event):
        """
        Regex form submit interaction option multi value matcher returning the matched group dictionaries as well.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        groups_and_values : `None | list<dict<str, str>, object)>`
            The matched values from the field's `custom_id` and their values.
        """
        groups_and_values = None
        
        for match, component_type, value_or_values in interaction_event.iter_matches_and_values(
            converter.annotation.fullmatch
        ):
            resolve = component_type.resolve
            if (resolve is None):
                continue
            
            if groups_and_values is None:
                groups_and_values = []
            
            groups_and_values.append((match.groupdict(), resolve(interaction_event.resolved, value_or_values)))
        
        return groups_and_values
    
    
    @staticmethod
    def _converter_regex_group_tuple(converter, interaction_event):
        """
        Regex form submit interaction option multi value matcher returning the matched group tuples as well.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        groups_and_values : `None | list<(tuple<str>`, object)>`
            The matched values from the field's `custom_id` and their values.
        """
        groups_and_values = None
        
        for match, component_type, value_or_values in interaction_event.iter_matches_and_values(
            converter.annotation.fullmatch
        ):
            resolve = component_type.resolve
            if (resolve is None):
                continue
            
            if (groups_and_values is None):
                groups_and_values = []
            
            groups_and_values.append((match.groups(), resolve(interaction_event.resolved, value_or_values)))
        
        return groups_and_values



from .client_platform_configuration import *
from .fields import *
from .preinstanced import *


__all__ = (
    *client_platform_configuration.__all__,
    *fields.__all__,
    *preinstanced.__all__,
)
